///////////////////////////////////////////////////////////////////////////////
// main.cpp
// ========
// camera with roations (pitch and yaw only) with shift (pan)
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2013-04-20
// UPDATED: 2024-03-22
///////////////////////////////////////////////////////////////////////////////

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <cstdlib>
#include <iostream>
#include <sstream>
#include <iomanip>
#include "Matrices.h"
#include "Vectors.h"
#include "ObjModel.h"
#include "Triangle.h"
#include "Edge.h"
#include "Line.h"
#include "Plane.h"

// GLUT CALLBACK functions
void displayCB();
void reshapeCB(int w, int h);
void timerCB(int millisec);
void idleCB();
void keyboardCB(unsigned char key, int x, int y);
void mouseCB(int button, int stat, int x, int y);
void mouseMotionCB(int x, int y);
void mousePassiveMotionCB(int x, int y);
void specialCB(int key, int x, int y);

// CALLBACK function when exit() called ///////////////////////////////////////
void exitCB();

void initGL();
int  initGLUT(int argc, char **argv);
bool initSharedMem();
void clearSharedMem();
void initLights();
void setCamera(float posX, float posY, float posZ, float targetX, float targetY, float targetZ);
void drawString(const char *str, int x, int y, float color[4], void *font);
void drawString3D(const char *str, float pos[3], float color[4], void *font);
void showInfo();
void toOrtho();
void toPerspective();
void drawAxis(float size=2.5f, float width=5.0f);
void drawGridXZ(float size=10.0f, float step=1.0f);
void drawGridXY(float size=10.0f, float step=1.0f);
void drawCamera();
void drawCameraAxis(float size=2.5f);
void drawCameraTarget(float size);
void drawCameraFocalLine();
void drawCameraLines(float x, float y, float z, float alpha=1.0f);
void drawCameraEdges(float angle=90.0f);
void drawCameraEdges2(float angle=90.0f);
void drawCenterLine(float size);
void drawModel();
std::vector<Triangle> generateTriangles();
void generateEdges();
int  isNewEdge(Edge edge);
void generateHardEdges(float angle);
void generateOutlineEdges(Matrix4 m);
void updateViewMatrix();
void shiftCamera(float deltaLeft, float deltaUp);
void forwardCamera(float delta);
void updateCamera();
void resetCamera();
Matrix4 setFrustum(float l, float r, float b, float t, float n, float f);
Matrix4 setFrustum(float fov, float aspect, float n, float f);
Matrix4 setOrthoFrustum(float l, float r, float b, float t, float n, float f);
void makePositiveZero(Vector3& vec);

// constants
const int   SCREEN_WIDTH    = 1200;
const int   SCREEN_HEIGHT   = 800;
const float CAMERA_DISTANCE = 10.0f;
const float FOV_Y           = 100.0f;
const int   TEXT_WIDTH      = 8;
const int   TEXT_HEIGHT     = 13;
const float DEG2RAD         = acos(-1) / 180.0f;
const float RAD2DEG         = 180.0f / acos(-1);

// global variables
void *font = GLUT_BITMAP_8_BY_13;
int screenWidth;
int screenHeight;
bool mouseLeftDown;
bool mouseRightDown;
float mouseX, mouseY;
int drawMode = 0;
ObjModel objCam;

// for edges
std::vector<Edge> edges;
std::vector<Edge> hardEdges;
std::vector<Edge> outlineEdges;

Matrix4 matrixPerspective;
Matrix4 matrixOrthographic;
Matrix4 matrixView1;        // 3rd-person view
Matrix4 matrixView2;        // 1st-person view
Matrix4 matrixCamera;       // camera model in world space

float cameraDistance;
Vector3 cameraPosition;
Vector3 cameraTarget;
Vector3 cameraAngle;        // (pitch, yaw, roll) in degree



///////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    // init global vars
    initSharedMem();

    // register exit callback
    atexit(exitCB);

    // init GLUT and GL
    initGLUT(argc, argv);
    initGL();

    // the last GLUT call (LOOP)
    // window will be shown and display callback is triggered by events
    // NOTE: this call never return main().
    glutMainLoop(); /* Start GLUT event-processing loop */

    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// draw a grid on XZ-plane
///////////////////////////////////////////////////////////////////////////////
void drawGridXZ(float size, float step)
{
    // disable lighting
    glDisable(GL_LIGHTING);

    glBegin(GL_LINES);

    glColor4f(0.5f, 0.4f, 0.5f, 0.5f);
    for(float i=step; i <= size; i+= step)
    {
        glVertex3f(-size, 0,  i);   // lines parallel to X-axis
        glVertex3f( size, 0,  i);
        glVertex3f(-size, 0, -i);   // lines parallel to X-axis
        glVertex3f( size, 0, -i);

        glVertex3f( i, 0, -size);   // lines parallel to Z-axis
        glVertex3f( i, 0,  size);
        glVertex3f(-i, 0, -size);   // lines parallel to Z-axis
        glVertex3f(-i, 0,  size);
    }

    // x-axis
    glColor3f(1, 0, 0);
    glVertex3f(-size, 0, 0);
    glVertex3f( size, 0, 0);

    // z-axis
    glColor3f(0,0,1);
    glVertex3f(0, 0, -size);
    glVertex3f(0, 0,  size);

    glEnd();

    // enable lighting back
    glEnable(GL_LIGHTING);
}



///////////////////////////////////////////////////////////////////////////////
// draw a grid on XY-plane
///////////////////////////////////////////////////////////////////////////////
void drawGridXY(float size, float step)
{
    // disable lighting
    glDisable(GL_LIGHTING);

    glBegin(GL_LINES);

    glColor4f(0.5f, 0.5f, 0.2f, 0.5f);
    for(float i=step; i <= size; i+= step)
    {
        glVertex3f(-size,  i, 0);   // lines parallel to X-axis
        glVertex3f( size,  i, 0);
        glVertex3f(-size, -i, 0);   // lines parallel to X-axis
        glVertex3f( size, -i, 0);

        glVertex3f( i, -size, 0);   // lines parallel to Y-axis
        glVertex3f( i,  size, 0);
        glVertex3f(-i, -size, 0);   // lines parallel to Y-axis
        glVertex3f(-i,  size, 0);
    }

    // x-axis
    glColor3f(1, 0, 0);
    glVertex3f(-size, 0, 0);
    glVertex3f( size, 0, 0);

    // y-axis
    glColor3f(0, 1, 0);
    glVertex3f(0, -size, 0);
    glVertex3f(0,  size, 0);

    glEnd();

    // enable lighting back
    glEnable(GL_LIGHTING);
}



///////////////////////////////////////////////////////////////////////////////
// draw the local axis of an object
///////////////////////////////////////////////////////////////////////////////
void drawAxis(float size, float width)
{
    glDepthFunc(GL_ALWAYS);     // to avoid visual artifacts with grid lines
    glDisable(GL_LIGHTING);

    // draw axis
    glLineWidth(width);
    glBegin(GL_LINES);
        glColor3f(1, 0, 0);
        glVertex3f(0, 0, 0);
        glVertex3f(size, 0, 0);
        glColor3f(0, 1, 0);
        glVertex3f(0, 0, 0);
        glVertex3f(0, size, 0);
        glColor3f(0, 0, 1);
        glVertex3f(0, 0, 0);
        glVertex3f(0, 0, size);
    glEnd();
    glLineWidth(1);

    // draw arrows(actually big square dots)
    glPointSize(width * 2);
    glBegin(GL_POINTS);
        glColor3f(1, 0, 0);
        glVertex3f(size, 0, 0);
        glColor3f(0, 1, 0);
        glVertex3f(0, size, 0);
        glColor3f(0, 0, 1);
        glVertex3f(0, 0, size);
    glEnd();
    glPointSize(1);

    // restore default settings
    glEnable(GL_LIGHTING);
    glDepthFunc(GL_LEQUAL);
}



///////////////////////////////////////////////////////////////////////////////
// draw a camera object
///////////////////////////////////////////////////////////////////////////////
void drawCamera()
{
    glColor4f(1.0f, 1.0f, 0.0f, 0.5f);

    // enable vertex arrays
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY);

    // before draw, specify vertex arrays
    const float* vertices = objCam.getInterleavedVertices();
    int stride = objCam.getInterleavedStride();
    glVertexPointer(3, GL_FLOAT, stride, vertices);
    glNormalPointer(GL_FLOAT, stride, vertices + 3);

    int count = objCam.getGroupCount();
    glCullFace(GL_FRONT);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    for(int i = 0; i < count; ++i)
    {
        glDrawElements(GL_TRIANGLES, objCam.getIndexCount(i), GL_UNSIGNED_INT, objCam.getIndices(i));
    }
    glCullFace(GL_BACK);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
    for(int i = 0; i < count; ++i)
    {
        glDrawElements(GL_TRIANGLES, objCam.getIndexCount(i), GL_UNSIGNED_INT, objCam.getIndices(i));
    }

    glDisableClientState(GL_VERTEX_ARRAY);  // disable vertex arrays
    glDisableClientState(GL_NORMAL_ARRAY);
}



///////////////////////////////////////////////////////////////////////////////
// draw edges with lines
///////////////////////////////////////////////////////////////////////////////
void drawCameraEdges(float angle)
{
    //float edgeColor[] = {1.0f, 0.9f, 0.0f, 1.0f};
    //float edgeColor[] = {0.6f, 0.5f, 0.2f, 1.0f};
    float edgeColor[] = {0.0f, 0.0f, 0.0f, 1.0f};

    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glLineWidth(1.0f);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glColor4fv(edgeColor);

    glBegin(GL_LINES);

    // draw hard edges
    int count = (int)hardEdges.size();
    for(int i = 0; i < count; ++i)
    {
        glVertex3fv(&hardEdges[i].vertex1[0]);
        glVertex3fv(&hardEdges[i].vertex2[0]);
    }

    // draw silhouette edges
    count = (int)outlineEdges.size();
    for(int i = 0; i < count; ++i)
    {
        glVertex3fv(&outlineEdges[i].vertex1[0]);
        glVertex3fv(&outlineEdges[i].vertex2[0]);
    }

    glEnd();

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
}
void drawCameraEdges2(float angle)
{
    //float edgeColor[] = {1.0f, 0.9f, 0.0f, 1.0f};
    //float edgeColor[] = {0.6f, 0.5f, 0.2f, 1.0f};
    float edgeColor[] = {0.0f, 0.0f, 0.0f, 1.0f};

    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glLineWidth(1.0f);
    //glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glColor4fv(edgeColor);

    glBegin(GL_LINES);

    // draw hard edges
    int count = (int)hardEdges.size();
    for(int i = 0; i < count; ++i)
    {
        glVertex3fv(&hardEdges[i].vertex1[0]);
        glVertex3fv(&hardEdges[i].vertex2[0]);
    }

    // draw silhouette edges
    count = (int)outlineEdges.size();
    for(int i = 0; i < count; ++i)
    {
        glVertex3fv(&outlineEdges[i].vertex1[0]);
        glVertex3fv(&outlineEdges[i].vertex2[0]);
    }

    glEnd();

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
}



///////////////////////////////////////////////////////////////////////////////
// draw the local axis of camera from current view matrix
///////////////////////////////////////////////////////////////////////////////
void drawCameraAxis(float size)
{
    //glDepthFunc(GL_ALWAYS);     // to avoid visual artifacts with grid lines
    glDisable(GL_LIGHTING);

    // draw axis
    glLineWidth(3);
    glBegin(GL_LINES);
        glColor3f(1, 0, 0);
        glVertex3f(0, 0, 0);
        glVertex3f(-size, 0, 0);
        glColor3f(0, 1, 0);
        glVertex3f(0, 0, 0);
        glVertex3f(0, size, 0);
        glColor3f(0, 0, 1);
        glVertex3f(0, 0, 0);
        glVertex3f(0, 0, -size);
    glEnd();
    glLineWidth(1);

    // draw arrows(actually big square dots)
    glPointSize(6);
    glBegin(GL_POINTS);
        glColor3f(1, 0, 0);
        glVertex3f(-size, 0, 0);
        glColor3f(0, 1, 0);
        glVertex3f(0, size, 0);
        glColor3f(0, 0, 1);
        glVertex3f(0, 0, -size);
    glEnd();
    glPointSize(1);

    // restore default settings
    glEnable(GL_LIGHTING);
    //glDepthFunc(GL_LEQUAL);
}



///////////////////////////////////////////////////////////////////////////////
// draw a point with a given size
///////////////////////////////////////////////////////////////////////////////
void drawCameraTarget(float size=1)
{
    glDepthFunc(GL_ALWAYS);     // to avoid visual artifacts with grid lines
    glDisable(GL_LIGHTING);

    glPointSize(size);
    glColor3f(1, 1, 0);

    glBegin(GL_POINTS);
    glVertex3f(cameraTarget.x, cameraTarget.y, cameraTarget.z);
    glEnd();

    glPointSize(1);

    // restore default settings
    glEnable(GL_LIGHTING);
    glDepthFunc(GL_LEQUAL);
}



///////////////////////////////////////////////////////////////////////////////
// draw camera's focal line
///////////////////////////////////////////////////////////////////////////////
void drawCameraFocalLine()
{
    Vector3 forward(-matrixView2[2], -matrixView2[6], -matrixView2[10]);
    //Vector3 forward(-matrixCamera[8], -matrixCamera[9], -matrixCamera[10]);
    Vector3 p = cameraPosition + (50 * forward);   // extend

    // draw focal guide lines
    glDisable(GL_LIGHTING);
    //glDisable(GL_DEPTH_TEST);

    glBegin(GL_LINES);
    glColor3f(1,1,0);
    glVertex3f(cameraPosition.x, cameraPosition.y, cameraPosition.z);
    glColor4f(1,1,0, 0);
    glVertex3f(p.x, p.y, p.z);
    glEnd();

    Line line(forward, cameraPosition);             // direction and point
    Plane plane(Vector3(0,0,1), Vector3(0,0,0));    // noraml and point
    Vector3 p2 = plane.intersect(line);

    glColor3f(1,1,0);
    glPointSize(6);
    glBegin(GL_POINTS);
    glVertex3f(p2.x, p2.y, p2.z);
    glEnd();

    glEnable(GL_LIGHTING);
    //glEnable(GL_DEPTH_TEST);
}



///////////////////////////////////////////////////////////////////////////////
// draw the center guide lines
///////////////////////////////////////////////////////////////////////////////
void drawCenterLine(float size)
{
    // backup current model-view matrix
    glMatrixMode(GL_MODELVIEW);         // switch to modelview matrix
    glPushMatrix();                     // save current modelview matrix
    glLoadIdentity();                   // reset modelview matrix

    // set to 2D orthogonal projection
    glMatrixMode(GL_PROJECTION);        // switch to projection matrix
    glPushMatrix();                     // save current projection matrix
    glLoadIdentity();                   // reset projection matrix
    glOrtho(0, screenWidth, 0, screenHeight, -1, 1); // set to orthogonal projection

    // center point in screen space
    float centerX = screenWidth * 0.5f;
    float centerY = screenHeight * 0.5f;

    // draw focal guide lines
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    //glLineWidth(1.0f);
    glColor3f(1,1,0);
    glBegin(GL_LINES);
    glVertex2f(centerX, centerY - size * 0.5f); // vertical line
    glVertex2f(centerX, centerY + size * 0.5f);
    glVertex2f(centerX - size * 0.5f, centerY); // horizontal line
    glVertex2f(centerX + size * 0.5f, centerY);
    glEnd();
    //glLineWidth(1);
    glEnable(GL_LIGHTING);
    glEnable(GL_DEPTH_TEST);

    glPopMatrix();                  // restore to previous projection matrix

    // restore modelview matrix
    glMatrixMode(GL_MODELVIEW);     // switch to modelview matrix
    glPopMatrix();                  // restore to previous modelview matrix
}



///////////////////////////////////////////////////////////////////////////////
// draw a model (tetrahedron)
///////////////////////////////////////////////////////////////////////////////
void drawModel()
{
    glColor3f(1, 1, 1);
    glBegin(GL_TRIANGLES);
    glNormal3f(0.6667f, 0.6667f, 0.3334f);
    glVertex3f(1, 0, 0);
    glVertex3f(0, 1, 0);
    glVertex3f(0, 0, 2);

    glNormal3f(-0.6667f, 0.6667f, 0.3334f);
    glVertex3f(-1, 0, 0);
    glVertex3f(0, 0, 2);
    glVertex3f(0, 1, 0);

    glNormal3f(0, 0, -1);
    glVertex3f(1, 0, 0);
    glVertex3f(0, 0, 2);
    glVertex3f(-1, 0, 0);

    glNormal3f(0, -1, 0);
    glVertex3f(1, 0, 0);
    glVertex3f(-1, 0, 0);
    glVertex3f(0, 1, 0);
    glEnd();
}



///////////////////////////////////////////////////////////////////////////////
// generate edges from triangles
///////////////////////////////////////////////////////////////////////////////
void generateEdges()
{
    Edge e[3];
    int edgeIndex = -1;
    edges.clear();

    std::vector<Triangle> triangles = generateTriangles();
    int triCount = (int)triangles.size();
    for(int i = 0; i < triCount; ++i)
    {
        e[0].vertex1 = triangles[i].vertex1;
        e[0].vertex2 = triangles[i].vertex2;
        e[0].normal1 = triangles[i].normal;

        e[1].vertex1 = triangles[i].vertex2;
        e[1].vertex2 = triangles[i].vertex3;
        e[1].normal1 = triangles[i].normal;

        e[2].vertex1 = triangles[i].vertex3;
        e[2].vertex2 = triangles[i].vertex1;
        e[2].normal1 = triangles[i].normal;

        for(int j = 0; j < 3; ++j)
        {
            edgeIndex = isNewEdge(e[j]);
            if(edgeIndex < 0)
            {
                edges.push_back(e[j]);
            }
            else
            {
                edges[edgeIndex].normal2 = triangles[i].normal;
                float dot = edges[edgeIndex].normal1.dot(edges[edgeIndex].normal2);
                if(dot < -1.0f)
                    dot = -1.0f;
                else if(dot > 1.0f)
                    dot = 1.0f;

                edges[edgeIndex].normalAngle = RAD2DEG * acosf(dot);
                //std::cout << i << ": " << edges[edgeIndex].normal1 << ", " << edges[edgeIndex].normal2 << ", " << edges[edgeIndex].normalAngle << std::endl;
            }
        }
    }

    std::cout << "Edge Count: " << edges.size() << std::endl;
}



///////////////////////////////////////////////////////////////////////////////
void generateHardEdges(float angle)
{
    hardEdges.clear();
    int count = (int)edges.size();
    for(int i = 0; i < count; ++i)
    {
        //std::cout << "Angle[" << i << "] " << edges[i].normalAngle << " degree\n";
        if(edges[i].normalAngle >= (angle - 0.1f))
        {
            hardEdges.push_back(edges[i]);
        }
    }
    std::cout << "Hard Edges Count: " << hardEdges.size() << std::endl;
}



///////////////////////////////////////////////////////////////////////////////
void generateOutlineEdges(Matrix4 m)
{
    Vector3 n1, n2;
    float a1, a2;
    Vector3 eye = Vector3(0,0,1);
    outlineEdges.clear();
    int count = (int)edges.size();
    for(int i = 0; i < count; ++i)
    {
        // transform face normals
        n1 = m * edges[i].normal1;
        n2 = m * edges[i].normal2;

        // get angles
        a1 = eye.dot(n1);
        a2 = eye.dot(n2);
        if((a1 * a2) < 0)
        {
            outlineEdges.push_back(edges[i]);
        }
    }
    //std::cout << "Outline Edges Count: " << outlineEdges.size() << std::endl;
}



///////////////////////////////////////////////////////////////////////////////
int isNewEdge(Edge e)
{
    int count = (int)edges.size();
    for(int i = 0; i < count; ++i)
    {
        if(e.vertex1 == edges[i].vertex2 && e.vertex2 == edges[i].vertex1)
            return i;
    }
    return -1;
}



///////////////////////////////////////////////////////////////////////////////
// get triangles
///////////////////////////////////////////////////////////////////////////////
std::vector<Triangle> generateTriangles()
{
    const float* vertices = objCam.getVertices();
    Vector3 v1, v2, v3;
    Triangle tri;
    std::vector<Triangle> triangles;

    int groupCount = objCam.getGroupCount();
    for(int i = 0; i < groupCount; ++i)
    {
        const unsigned int* indices = objCam.getIndices(i);
        int indexCount = objCam.getIndexCount(i);
        for(int j = 0; j < indexCount; j += 3)
        {
            int k = indices[j] * 3;
            tri.vertex1.set(vertices[k], vertices[k+1], vertices[k+2]);
            k = indices[j+1] * 3;
            tri.vertex2.set(vertices[k], vertices[k+1], vertices[k+2]);
            k = indices[j+2] * 3;
            tri.vertex3.set(vertices[k], vertices[k+1], vertices[k+2]);

            // compute face normal
            tri.normal = (tri.vertex2 - tri.vertex1).cross(tri.vertex3 - tri.vertex1);
            tri.normal.normalize();

            //std::cout << tri.vertex1 << "," << tri.vertex2 << "," << tri.vertex3 << std::endl;
            triangles.push_back(tri);
        }
    }
    std::cout << "triangles: " << triangles.size() << std::endl;
    return triangles;
}



///////////////////////////////////////////////////////////////////////////////
// initialize GLUT for windowing
///////////////////////////////////////////////////////////////////////////////
int initGLUT(int argc, char **argv)
{
    // GLUT stuff for windowing
    // initialization openGL window.
    // it is called before any other GLUT routine
    glutInit(&argc, argv);

    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);   // display mode

    glutInitWindowSize(screenWidth, screenHeight);  // window size

    glutInitWindowPosition(100, 100);           // window location

    // finally, create a window with openGL context
    // Window will not displayed until glutMainLoop() is called
    // it returns a unique ID
    int handle = glutCreateWindow(argv[0]);     // param is the title of window

    // register GLUT callback functions
    glutDisplayFunc(displayCB);
    glutTimerFunc(33, timerCB, 33);             // redraw only every given millisec
    //glutIdleFunc(idleCB);                       // redraw whenever system is idle
    glutReshapeFunc(reshapeCB);
    glutKeyboardFunc(keyboardCB);
    //glutMouseFunc(mouseCB);
    //glutMotionFunc(mouseMotionCB);
    glutPassiveMotionFunc(mousePassiveMotionCB);
    glutSpecialFunc(specialCB);

    return handle;
}



///////////////////////////////////////////////////////////////////////////////
// initialize OpenGL
// disable unused features
///////////////////////////////////////////////////////////////////////////////
void initGL()
{
    glShadeModel(GL_SMOOTH);                    // shading mathod: GL_SMOOTH or GL_FLAT
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);      // 4-byte pixel alignment

    // enable /disable features
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_CULL_FACE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_SCISSOR_TEST);

     // track material ambient and diffuse from surface color, call it before glEnable(GL_COLOR_MATERIAL)
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);

    glClearColor(0, 0, 0, 1.0f);                // background color
    glClearStencil(0);                          // clear stencil buffer
    glClearDepth(1.0f);                         // 0 is near, 1 is far
    glDepthFunc(GL_LEQUAL);

    initLights();
}



///////////////////////////////////////////////////////////////////////////////
// write 2d text using GLUT
// The projection matrix must be set to orthogonal before call this function.
///////////////////////////////////////////////////////////////////////////////
void drawString(const char *str, int x, int y, float color[4], void *font)
{
    glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT); // lighting and color mask
    glDisable(GL_LIGHTING);     // need to disable lighting for proper text color
    glDisable(GL_TEXTURE_2D);

    glColor4fv(color);          // set text color
    glRasterPos2i(x, y);        // place text position

    // loop all characters in the string
    while(*str)
    {
        glutBitmapCharacter(font, *str);
        ++str;
    }

    glEnable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
    glPopAttrib();
}



///////////////////////////////////////////////////////////////////////////////
// draw a string in 3D space
///////////////////////////////////////////////////////////////////////////////
void drawString3D(const char *str, float pos[3], float color[4], void *font)
{
    glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT); // lighting and color mask
    glDisable(GL_LIGHTING);     // need to disable lighting for proper text color
    glDisable(GL_TEXTURE_2D);

    glColor4fv(color);          // set text color
    glRasterPos3fv(pos);        // place text position

    // loop all characters in the string
    while(*str)
    {
        glutBitmapCharacter(font, *str);
        ++str;
    }

    glDisable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
    glPopAttrib();
}



///////////////////////////////////////////////////////////////////////////////
// initialize global variables
///////////////////////////////////////////////////////////////////////////////
bool initSharedMem()
{
    screenWidth = SCREEN_WIDTH;
    screenHeight = SCREEN_HEIGHT;

    mouseLeftDown = mouseRightDown = false;
    mouseX = mouseY = 0;

    // load camera model
    objCam.read("camera.obj");
    objCam.printSelf();

    // generating polygons and edges
    generateEdges();
    generateHardEdges(70);

    // init camera
    resetCamera();

    // 3rd-person view matrix
    matrixView1.identity();
    matrixView1.translate(2, 3, 0);
    matrixView1.rotateY(-35);
    matrixView1.rotateX(30);
    matrixView1.translate(0, 0, -15);

    // init perspective matrix
    matrixPerspective = setFrustum(FOV_Y, (float)screenWidth / screenHeight, 1, 1000);

    drawMode = 0; // 0:fill, 1: wireframe, 2:points

    return true;
}



///////////////////////////////////////////////////////////////////////////////
// clean up shared memory
///////////////////////////////////////////////////////////////////////////////
void clearSharedMem()
{
}



///////////////////////////////////////////////////////////////////////////////
// initialize lights
///////////////////////////////////////////////////////////////////////////////
void initLights()
{
    // set up light colors (ambient, diffuse, specular)
    GLfloat lightKa[] = {.2f, .2f, .2f, 1.0f};  // ambient light
    GLfloat lightKd[] = {.7f, .7f, .7f, 1.0f};  // diffuse light
    GLfloat lightKs[] = {1, 1, 1, 1};           // specular light
    glLightfv(GL_LIGHT0, GL_AMBIENT, lightKa);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, lightKd);
    glLightfv(GL_LIGHT0, GL_SPECULAR, lightKs);

    // position the light
    float lightPos[4] = {0, 0, 20, 1}; // positional light
    glLightfv(GL_LIGHT0, GL_POSITION, lightPos);

    glEnable(GL_LIGHT0);                        // MUST enable each light source after configuration
}



///////////////////////////////////////////////////////////////////////////////
// set camera position and lookat direction
///////////////////////////////////////////////////////////////////////////////
void setCamera(float posX, float posY, float posZ, float targetX, float targetY, float targetZ)
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(posX, posY, posZ, targetX, targetY, targetZ, 0, 1, 0); // eye(x,y,z), focal(x,y,z), up(x,y,z)
}



///////////////////////////////////////////////////////////////////////////////
// display info messages
///////////////////////////////////////////////////////////////////////////////
void showInfo()
{
    // backup current model-view matrix
    glPushMatrix();                     // save current modelview matrix
    glLoadIdentity();                   // reset modelview matrix

    // set to 2D orthogonal projection
    glMatrixMode(GL_PROJECTION);        // switch to projection matrix
    glPushMatrix();                     // save current projection matrix
    glLoadIdentity();                   // reset projection matrix
    gluOrtho2D(0, screenWidth, 0, screenHeight); // set to orthogonal projection

    float color[4] = {1, 1, 1, 1};

    std::stringstream ss;
    ss << std::fixed << std::setprecision(3);

    // print camera info
    ss << "Mouse: (" << (int)mouseX << ", " << (int)mouseY << ")";
    drawString(ss.str().c_str(), 1, screenHeight - TEXT_HEIGHT, color, font);

    ss.str("");
    ss << "===== CAMERA =====";
    drawString(ss.str().c_str(), 1, screenHeight - 2*TEXT_HEIGHT, color, font);

    ss.str("");
    ss << "Position: (" << cameraPosition.x << ", " << cameraPosition.y << ", " << cameraPosition.z << ")";
    drawString(ss.str().c_str(), 1, screenHeight - 3*TEXT_HEIGHT, color, font);

    ss.str("");
    ss << "  Target: (" << cameraTarget.x << ", " << cameraTarget.y << ", " << cameraTarget.z << ")";
    drawString(ss.str().c_str(), 1, screenHeight - 4*TEXT_HEIGHT, color, font);

    ss.str("");
    ss << "    Left: (" << -matrixView2[0] << ", " << -matrixView2[4] << ", " << -matrixView2[8] << ")";
    drawString(ss.str().c_str(), 1, screenHeight - 5*TEXT_HEIGHT, color, font);

    ss.str("");
    ss << "      Up: (" << matrixView2[1] << ", " << matrixView2[5] << ", " << matrixView2[9] << ")";
    drawString(ss.str().c_str(), 1, screenHeight - 6*TEXT_HEIGHT, color, font);

    ss.str("");
    ss << " Forward: (" << -matrixView2[2] << ", " << -matrixView2[6] << ", " << -matrixView2[10] << ")";
    drawString(ss.str().c_str(), 1, screenHeight - 7*TEXT_HEIGHT, color, font);

    ss.str("");
    ss << "   Pitch: " << cameraAngle.x << " (X-axis rotation)";
    drawString(ss.str().c_str(), 1, screenHeight - 8*TEXT_HEIGHT, color, font);

    ss.str("");
    ss << "     Yaw: " << cameraAngle.y << " (Y-axis rotation)";
    drawString(ss.str().c_str(), 1, screenHeight - 9*TEXT_HEIGHT, color, font);

    ss.str("");
    ss << "Use arrow keys to shift camera and +/- to move forward.";
    drawString(ss.str().c_str(), 1, 1, color, font);

    // unset floating format
    ss << std::resetiosflags(std::ios_base::fixed | std::ios_base::floatfield);

    // restore projection matrix
    glPopMatrix();                   // restore to previous projection matrix

    // restore modelview matrix
    glMatrixMode(GL_MODELVIEW);      // switch to modelview matrix
    glPopMatrix();                   // restore to previous modelview matrix
}



///////////////////////////////////////////////////////////////////////////////
// set projection matrix as orthogonal
///////////////////////////////////////////////////////////////////////////////
void toOrtho()
{
    // set viewport to be the entire window
    glViewport(0, 0, (GLsizei)screenWidth, (GLsizei)screenHeight);

    // set orthographic viewing frustum
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, screenWidth, 0, screenHeight, -1, 1);

    // switch to modelview matrix in order to set scene
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}



///////////////////////////////////////////////////////////////////////////////
// set the projection matrix as perspective
///////////////////////////////////////////////////////////////////////////////
void toPerspective()
{
    // set viewport to be the entire window
    glViewport(0, 0, (GLsizei)screenWidth, (GLsizei)screenHeight);

    // set perspective viewing frustum
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0f, (float)(screenWidth)/screenHeight, 1.0f, 1000.0f); // FOV, AspectRatio, NearClip, FarClip

    // switch to modelview matrix in order to set scene
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}



///////////////////////////////////////////////////////////////////////////////
// equivalent to glFrustum()
// PARAMS: (left, right, bottom, top, near, far)
///////////////////////////////////////////////////////////////////////////////
Matrix4 setFrustum(float l, float r, float b, float t, float n, float f)
{
    Matrix4 mat;
    mat[0]  = 2 * n / (r - l);
    mat[5]  = 2 * n / (t - b);
    mat[8]  = (r + l) / (r - l);
    mat[9]  = (t + b) / (t - b);
    mat[10] = -(f + n) / (f - n);
    mat[11] = -1;
    mat[14] = -(2 * f * n) / (f - n);
    mat[15] = 0;
    return mat;
}



///////////////////////////////////////////////////////////////////////////////
// equivalent to gluPerspective()
// PARAMS: (vertical FOV, aspect ratio = width/height, near, far)
///////////////////////////////////////////////////////////////////////////////
Matrix4 setFrustum(float fovY, float aspect, float front, float back)
{
    float tangent = tanf(fovY/2 * DEG2RAD); // tangent of half fovY
    float height = front * tangent;         // half height of near plane
    float width = height * aspect;          // half width of near plane

    // params: left, right, bottom, top, near, far
    return setFrustum(-width, width, -height, height, front, back);
}



///////////////////////////////////////////////////////////////////////////////
// equivalent to glOrtho()
// PARAMS: (left, right, bottom, top, near, far)
///////////////////////////////////////////////////////////////////////////////
Matrix4 setOrthoFrustum(float l, float r, float b, float t, float n, float f)
{
    Matrix4 mat;
    mat[0]  = 2 / (r - l);
    mat[5]  = 2 / (t - b);
    mat[10] = -2 / (f - n);
    mat[12] = -(r + l) / (r - l);
    mat[13] = -(t + b) / (t - b);
    mat[14] = -(f + n) / (f - n);
    return mat;
}



///////////////////////////////////////////////////////////////////////////////
// update the view matrix for the 3rd-person view
///////////////////////////////////////////////////////////////////////////////
void updateViewMatrix()
{
    // construct view matrix
    matrixView1.identity();
}



///////////////////////////////////////////////////////////////////////////////
// update the matrix for 3rd-person view and the camera object
///////////////////////////////////////////////////////////////////////////////
void updateCamera()
{
    // construct matrix for camera object
    // NOTE: this camera lens is facing to -Z axis, so apply x & z rotations
    // are opposite direction
    matrixCamera.identity();
    matrixCamera.rotateZ(-cameraAngle.z);
    matrixCamera.rotateY(cameraAngle.y);
    matrixCamera.rotateX(-cameraAngle.x);
    matrixCamera.translate(cameraPosition);

    // construct 1st-person view matrix
    // NOTE: Y-axis in eye-space is same as world-space,
    // but, X and Z axis are opposite in eye-space, so reverse Y rotation only
    matrixView2.identity();
    matrixView2.translate(-cameraPosition);
    matrixView2.rotateZ(cameraAngle.z);     // roll
    matrixView2.rotateY(-cameraAngle.y);    // heading (yaw)
    matrixView2.rotateX(cameraAngle.x);     // pitch

    // re-compute camera target position in world space
    Vector3 forward(-matrixView2[2], -matrixView2[6], -matrixView2[10]);
    cameraTarget = cameraPosition + (cameraDistance * forward);

}



///////////////////////////////////////////////////////////////////////////////
// shift camera left/right/up/down in world space
// it will update the camera's position and target point
///////////////////////////////////////////////////////////////////////////////
void shiftCamera(float deltaLeft, float deltaUp)
{
    // left/up/forward vectors are 3 row vectors from the view matrix because
    // view matrix contains inverted rotations (transposed)
    //      Left = Vector3(-m[0], -m[4], -m[8])
    //        Up = Vector3( m[1],  m[5],  m[9])
    //   Forward = Vector3(-m[2], -m[6], -m[10])
    Vector3 cameraLeft(-matrixView2[0], -matrixView2[4], -matrixView2[8]);
    Vector3 cameraUp(matrixView2[1], matrixView2[5], matrixView2[9]);
    Vector3 cameraForward(-matrixView2[2], -matrixView2[6], -matrixView2[10]);

    // compute delta shift
    Vector3 deltaMovement = deltaLeft * cameraLeft; // move horizontal
    deltaMovement += deltaUp * cameraUp;            // add vertical

    // update camera's position by adding delta shift
    cameraPosition += deltaMovement;

    // update camera's target point using forward vector
    cameraTarget = cameraPosition + (cameraDistance * cameraForward);

    //DEBUG
    //std::cout << "delta: " << deltaMovement << std::endl;
}



///////////////////////////////////////////////////////////////////////////////
// move camera forward/backward in world space
// it will update the camera's position and target point
///////////////////////////////////////////////////////////////////////////////
void forwardCamera(float delta)
{
    // left/up/forward vectors are 3 row vectors from the view matrix because
    // view matrix contains inverted rotations (transposed)
    //      Left = Vector3(-m[0], -m[4], -m[8])
    //        Up = Vector3( m[1],  m[5],  m[9])
    //   Forward = Vector3(-m[2], -m[6], -m[10])
    Vector3 cameraForward(-matrixView2[2], -matrixView2[6], -matrixView2[10]);

    // compute delta shift
    Vector3 deltaMovement = cameraForward * delta;

    // update camera's position by adding delta movement
    cameraPosition += deltaMovement;

    // update camera's target point using forward vector
    cameraTarget = cameraPosition + (cameraDistance * cameraForward);
}



///////////////////////////////////////////////////////////////////////////////
// reset camera to default
///////////////////////////////////////////////////////////////////////////////
void resetCamera()
{
    cameraDistance = CAMERA_DISTANCE;
    cameraPosition.set(0, 0, CAMERA_DISTANCE);
    cameraTarget.set(0, 0, 0);
    cameraAngle.set(0, 0, 0);
    updateCamera();
}













//=============================================================================
// CALLBACKS
//=============================================================================

void displayCB()
{
    // clear whole screen
    glViewport(0, 0, screenWidth, screenHeight);
    glScissor(0, 0, screenWidth, screenHeight);
    glClearColor(0, 0, 0, 1.0f);                // background color
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    // 1st-person view ====================================
    // set perspective projection
    glMatrixMode(GL_PROJECTION);
    matrixPerspective = setFrustum(FOV_Y, (float)screenWidth / screenHeight, 1, 1000);
    glLoadMatrixf(matrixPerspective.get());

    // switch to modelview matrix in order to set scene
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf(matrixView2.get());

    // draw scene
    drawAxis();
    drawGridXY(20);
    //drawGridXZ(20);
    drawCameraTarget(10);                // camera target in world space
    drawCenterLine(100);

    // draw info messages
    showInfo();

    // 3rd-person view ====================================
    int overlayW = 400;
    int overlayH = 250;
    glViewport(13, 13, overlayW, overlayH);
    glScissor(13, 13, overlayW, overlayH);
    glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    // set perspective projection
    glMatrixMode(GL_PROJECTION);
    matrixPerspective = setFrustum(60.0f, (float)overlayW / overlayH, 1, 1000);
    glLoadMatrixf(matrixPerspective.get());

    // switch to modelview matrix in order to set scene
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf(matrixView1.get());

    // draw scene
    drawAxis(2.5f, 3.0f);
    drawGridXY(20);
    //drawGridXZ(20);
    drawCameraFocalLine();

    // draw camera object at world space
    Matrix4 matrix = matrixView1 * matrixCamera;
    glLoadMatrixf(matrix.get());
    drawCamera();
    drawCameraEdges2();
    drawCameraAxis(1);

    glutSwapBuffers();
}



void reshapeCB(int w, int h)
{
    screenWidth = w;
    screenHeight = h;
    //matrixPerspective = setFrustum(FOV_Y, (float)screenWidth / screenHeight, 1, 1000);

}


void timerCB(int millisec)
{
    glutTimerFunc(millisec, timerCB, millisec);
    glutPostRedisplay();
}


void idleCB()
{
    glutPostRedisplay();
}


void keyboardCB(unsigned char key, int x, int y)
{
    switch(key)
    {
    case 27: // ESCAPE
        exit(0);
        break;

    case ' ':
        break;

    case 'r':
    case 'R':
        resetCamera();
        break;

    case 'z':
    case 'Z':
        cameraAngle.z += 1.0f;
        updateCamera();
        break;

    case 'x':
    case 'X':
        cameraAngle.z -= 1.0f;
        updateCamera();
        break;

    case '+':
    case '=':
        forwardCamera(0.1f);
        updateCamera();
        break;

    case '-':
    case '_':
        forwardCamera(-0.1f);
        updateCamera();
        break;

    case 'd': // switch rendering modes (fill -> wire -> point)
    case 'D':
        drawMode = ++drawMode % 3;
        if(drawMode == 0)        // fill mode
        {
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            glEnable(GL_DEPTH_TEST);
            glEnable(GL_CULL_FACE);
        }
        else if(drawMode == 1)  // wireframe mode
        {
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_CULL_FACE);
        }
        else                    // point mode
        {
            glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_CULL_FACE);
        }
        break;

    default:
        ;
    }
}


void mouseCB(int button, int state, int x, int y)
{
    mouseX = x;
    mouseY = y;

    if(button == GLUT_LEFT_BUTTON)
    {
        if(state == GLUT_DOWN)
            mouseLeftDown = true;
        else if(state == GLUT_UP)
            mouseLeftDown = false;
    }

    else if(button == GLUT_RIGHT_BUTTON)
    {
        if(state == GLUT_DOWN)
            mouseRightDown = true;
        else if(state == GLUT_UP)
            mouseRightDown = false;
    }
}


void mouseMotionCB(int x, int y)
{
    if(mouseLeftDown)
    {
        cameraAngle.y += (x - mouseX);
        cameraAngle.x += (y - mouseY);
        mouseX = x;
        mouseY = y;
    }
    if(mouseRightDown)
    {
        cameraDistance -= (y - mouseY) * 0.2f;
        mouseY = y;
    }

    // re-compute view matrix for left-side
    //updateCamera();
}


void mousePassiveMotionCB(int x, int y)
{
    // delta from the center of screen
    //NOTE: rotate to left is positive yaw(Y-axis) angle and
    //      rotate to right is negative yaw angle for the camera
    //      rotate to down is positive pitch (X-axis) angle and
    //      rotate to up is negative pich for the camaera
    cameraAngle.y = -(x - screenWidth * 0.5f) * 0.1f;   // yaw (heading)
    cameraAngle.x = (y - screenHeight * 0.5f) * 0.1f;   // pitch
    mouseX = x;
    mouseY = y;
    updateCamera();
}


void specialCB(int key, int x, int y)
{
    const float DELTA = 0.1f;
    Vector2 delta;
    switch(key)
    {
    case GLUT_KEY_LEFT:
        delta.x = DELTA;            // shift camera to left
        break;

    case GLUT_KEY_RIGHT:
        delta.x = -DELTA;           // shift camera to right
        break;

    case GLUT_KEY_UP:
        delta.y = DELTA;            // shift camera upward
        break;

    case GLUT_KEY_DOWN:
        delta.y = -DELTA;           // shiftcamera downward
        break;

    default:
        ;
    }

    // pan camera to adjust camera's position and target points
    shiftCamera(delta.x, delta.y);  // left, up

    // re-compute camera
    updateCamera();
}


void exitCB()
{
    clearSharedMem();
}
